/* We illustrate the operation of the code using the problem

        min \sum_{i=1}^n  exp (x [i]) - sqrt (i)*x [i] ;
        subject to 0 <= x <= 1

   We solve the problem 4 times using different parameter setting.
   The 1st problem was solved using Hessian information (through myHessSolve)
   The rest 3 problems were solved only using function values and gradients.   
   On a linux workstation, the final statistics of the 1st solve generated 
   by driver was the following:
   
Final convergence status = 0
Convergence tolerance for gradient satisfied
projected gradient max norm:  6.963834e-10
function value:              -4.011035e+02

Total cg  iterations:                    3
Total cg  function evaluations:          4
Total cg  gradient evaluations:          3
Total cbb iterations:                    3
Total cbb function evaluations:          4
Total cbb gradient evaluations:          4
------------------------------------------
Total function evaluations:              8
Total gradient evaluations:              7
==========================================  

   the final statistics of the rest 3 solve generated by driver
   were as follows:

Final convergence status = 0
Convergence tolerance for gradient satisfied
projected gradient max norm:  2.618104e-10
function value:              -4.011035e+02

Total cg  iterations:                    8
Total cg  function evaluations:          9
Total cg  gradient evaluations:         15
Total cbb iterations:                    3
Total cbb function evaluations:          4
Total cbb gradient evaluations:          4
------------------------------------------
Total function evaluations:             13
Total gradient evaluations:             19
==========================================

The principal change in Version3.0 is to update the conjugate
gradient routine to CG_DESCENT 6.0. Note that the objective
structure also contains a pointer to the array of free indices 

The principal change in Version4.0 is to allow the user to
use second order Hessian information to accelerate convergence */

#include <math.h>
#include <stdio.h>
#include "asa_user.h" /* needed by the program which calls asa_cg*/

/* prototypes for the function or gradient evaluation routines  */
double myvalue
(
    asa_objective *asa
) ;

void mygrad
(
    asa_objective *asa
) ;

double myvalgrad
(
    asa_objective *asa
) ;

/* prototypes for solving A*d = b or calculate d = A*b, 
   where  A = Hess(ifree,ifree) + Hscale I, with Hscale >= 0. */  
void myHessSolve
(
    asa_objective *asa
) ;

int main (void)
{
    double *x, *lo, *hi ;
    INT i, n ;

    /* if you want to change parameter value, you need the following: */
    asacg_parm cgParm ;
    asa_parm asaParm ;

    /* allocate arrays for problem solution and bounds */
    n = 100 ; /* problem dimension */
    x  = (double *) malloc (n*sizeof (double)) ;
    lo = (double *) malloc (n*sizeof (double)) ;
    hi = (double *) malloc (n*sizeof (double)) ;
    for (i = 0; i < n; i++) lo [i] = (double) 0 ;
    for (i = 0; i < n; i++) hi [i] = (double) 1 ;

    /* if you want to change parameter value, initialize strucs with default */
    asa_cg_default (&cgParm) ;
    asa_default (&asaParm) ;

    /* if you want to change parameters, change them here: */
    cgParm.PrintParms = 0 ;
    cgParm.PrintLevel = 0 ;
    asaParm.PrintParms = 0;
    asaParm.PrintLevel = 0 ;
    asaParm.PrintFinal = 1 ;

    /* starting guess */
    for (i = 0; i < n; i++) x [i] = 1 ;

    /* run the code using both first order and second order (Hessian) information  */
    asa_cg (x, lo, hi, n, NULL, &cgParm, &asaParm,
                     1.e-8, myvalue, mygrad, myvalgrad, myHessSolve, NULL, NULL) ;

    /* if myHessSolve is not available or prefer not using Hessian, set myHessSolve = NULL */
    for (i = 0; i < n; i++) x [i] = 1 ; /* starting guess */
    asa_cg (x, lo, hi, n, NULL, &cgParm, &asaParm,
                     1.e-8, myvalue, mygrad, myvalgrad, NULL, NULL, NULL) ;
                     
    /* if no change in parameters, you could replace Parm arguments by NULL*/
    for (i = 0; i < n; i++) x [i] = 1 ; /* starting guess */
    asa_cg (x, lo, hi, n, NULL, NULL, NULL,
                     1.e-8, myvalue, mygrad, myvalgrad, NULL, NULL, NULL) ;

    /* with some loss of efficiency, you could omit the valgrad routine */
    for (i = 0; i < n; i++) x [i] = 1 ; /* starting guess */
    asa_cg (x, lo, hi, n, NULL, NULL, NULL, 1.e-8, myvalue, mygrad, NULL, NULL,
            NULL, NULL);

    free (x) ;
    free (lo) ;
    free (hi) ;
}

double myvalue /* evaluate the objective function */
(
    asa_objective *asa
)
{
    double f, t, *x ;
    INT i, n ;
    x = asa->x ;
    n = asa->n ;
    f = 0. ;
    for (i = 0; i < n; i++)
    {
        t = i + 1 ;
        t = sqrt (t) ;
        f += exp (x [i]) - t*x [i] ;
    }
    return (f) ;
}

void mygrad /* evaluate the gradient of the objective function */
(
    asa_objective *asa
)
{
    double t, *g, *x ;
    INT i, n ;
    x = asa->x ;
    g = asa->g ;
    n = asa->n ;
    for (i = 0; i < n; i++)
    {
        t = i + 1 ;
        t = sqrt (t) ;
        g [i] = exp (x [i]) -  t ;
    }
    return ;
}

double myvalgrad /* value and gradient of the objective function */
(
    asa_objective *asa
)
{
    double f, xi, t, *g, *x ;
    INT i, n ;
    x = asa->x ;
    g = asa->g ;
    n = asa->n ;
    f = 0 ;
    for (i = 0; i < n; i++)
    {
        t = i + 1 ;
        t = sqrt (t) ;
        xi = x [i] ;
        f += exp (xi) - t*xi ;
        g [i] = exp (xi) -  t ;
    }
    return (f) ;
}


/* Hinfo = 1 => evaluate Hess at x and solve A*d = b;
   Hinfo = 2 => use previous Hess and solve A*d = b ;
   Hinfo = 3 => evaluate Hess at x and calculate d = A*b;
   Hinfo = 4 => use previous Hess and calculate d = A*b ;                
where  A = Hess(ifree,ifree) + Hscale I, with Hscale >= 0.   
*/   
void myHessSolve
(
    asa_objective *asa
)
{
    double Hscale, *x, *b, *d ;
    int Hinfo ;
    INT i, j, n, nfree, *ifree ;
    
    x = asa->x ;
    b = asa->b ;
    d = asa->d ;
    Hscale = asa->Hscale ;
    ifree = asa->ifree ;
    nfree = asa->nfree ;
    Hinfo = asa->Hinfo ;
    n = asa->n ;
    
    // Note that nfree == n => ifree = NULL !!!
    
    /* Let A = Hess(ifree,ifree) + Hscale I */ 
    if ( Hinfo == 1 )  /* evaluate Hess at x and solve A*d = b */
    {
        if ( nfree == n )
        {
            for (i = 0; i < n; i++)
            {
                d [i] = b [i] / (exp (x[i]) + Hscale) ;            
            }                
        }
        else
        {
            for (j = 0; j < nfree; j++)
            {
                d [j] = b [j] / (exp (x[ifree[j]]) + Hscale) ;
            }  
        }            
    }
    else if ( Hinfo == 2 )  /* use previous Hess and solve A*d = b */
    {
        if ( nfree == n )
        {
            for (i = 0; i < n; i++)
            {
                d [i] = b [i] / (exp (x[i]) + Hscale) ;
            }                
            
        }
        else
        {
            for (j = 0; j < nfree; j++)
            {
                d [j] = b [j] / (exp (ifree[j]) + Hscale) ;
            }  
        }        
    }
    else if ( Hinfo == 3 )  /* evaluate Hess at x and calculate d = A*b */
    {    
        if ( nfree == n )
        {
            for (i = 0; i < n; i++)
            {
                d [i] = (exp (x[i])+ Hscale)*b [i] ;
            }                
            
        } 
        else
        {                
            for (j = 0; j < nfree; j++)
            {
                d [j] = (exp (x[ifree[j]])+ Hscale)*b [j] ;
            }
        }
    }
    else  /* Hinfo =4, use previous Hess and calculate d = A*b */
    {
        if ( nfree == n )
        {
            for (i = 0; i < n; i++)
            {
                d [i] = (exp (x[i])+ Hscale)*b [i] ;
            }                
            
        } 
        else
        {                
            for (j = 0; j < nfree; j++)
            {
                d [j] = (exp (x[ifree[j]])+ Hscale)*b [j] ;
            }
        }             
    }
}

